# Pingback.pm -- handle pings and pingbacks for RoBlog
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
# more details.
# 
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307  USA.
#
package RoBlog::Pingback;

use Encode;
use LWP::UserAgent;
use HTML::Entities;
use Apache2::Const;
use Apache2::Request;
use POSIX;
use RoBlog::Util;
use RoBlog::Config;
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(send_ping send_pingback recv_pingback send_ping);

no warnings 'redefine';
use strict;

sub send_ping {
	my $answer = "";

	foreach my $url (@pingurls) {
		my $ua = new LWP::UserAgent();
		my $req = new HTTP::Request("POST", $url);
		$req->content_type("text/xml");
		$req->add_content(<<"EOF");
<?xml version="1.0"?>
<methodCall>
  <methodName>weblogUpdates.ping</methodName>
  <params>
    <param><value><string>$blogtitle</string></value></param>
    <param><value><string>$baseurl</string></value></param>
  </params>
</methodCall>
EOF
		my $resp = $ua->request($req);
		my $reply = $resp->content();
		if (!$reply) {
			$answer .= vmsg("Failed to ping $url: ".$resp->status_line());
			next;
		}
		my $entry = xmlparse($reply, "methodResponse", [], [qw(member)]);
		my $members = $entry->{params}{param}{value}{struct}{member};
		if (!$members) {
			$answer .= vmsg("Failed to ping $url: no structure response");
			next;
		}
		if (xmlmember($members, "flerror") ne "0") {
			$answer .= vmsg("Failed to ping $url: ".
							xmlmember($members, "message"));
		}
		else {
			vmsg("Successfully pinged $url: ".
				 (xmlmember($members, "message") || "(no message)"));
		}
	}
	return $answer;
}

sub send_pingback {
	my ($from, $to) = @_;

    # fetch the page
	my $ua = new LWP::UserAgent();
    my $req = new HTTP::Request("GET", $to);
    $req->referer($from);
	$ua->max_size(4096);
    my $resp = $ua->request($req);
	return vmsg("Failed to fetch $to: ".$resp->status_line())
		if !$resp->is_success();
    # scan for a pingback link
    my $server;
    if (!($server = ($resp->header("X-Pingback"))[0])) {
		$server = decode_entities($))
			if $resp->content() =~
			m,<link\s+rel=\"pingback\"\s+href=\"([^\"]+)\"\s*/?>,s; #"
    }
    return vmsg("No pingback server found for $to") if !$server;
    # send pingback
	$req = new HTTP::Request("POST", $server);
	$req->content_type("text/xml; charset=utf-8");
	$req->add_content(<<"EOF");
<?xml version="1.0"?>
<methodCall>
  <methodName>pingback.ping</methodName>
  <params>
    <param><value><string>$from</string></value></param>
    <param><value><string>$to</string></value></param>
  </params>
</methodCall>
EOF
	$resp = $ua->request($req);
	my $reply = $resp->content();
	return vmsg("Failed to post to $server: ".$resp->status_line()) if !$reply;
	my $xmlresp = xmlparse($reply, "methodResponse", [], [qw(member)]);
	return vmsg("Failed to post to $server: XML parse error: $1")
		if $xmlresp =~ /^ERROR ((.|\n)*)$/;
	return vmsg("Pingback call to $server failed: ".
				xmlmember($xmlresp->{fault}{value}{struct}{members}, "faultString"))
		if exists($xmlresp->{fault});
	vmsg("Successfully sent pingback to $server");
	return undef;
}

sub recv_pingback {
	my ($req, $dbh) = @_;

	$req->method() eq "POST"
		or return xmlerror("405 Method not allowed",
						   -32300, "Only XML-RPC POST requests recognised.");
	my $xml = "";
	{ local($/) = undef; $xml = <STDIN>; }
	my $call = xmlparse($xml, "methodCall", [], [qw(param)]);
	return xmlerror("400 Bad request", -32700, "Parse error: $1")
		if $call =~ /^ERROR ((.|\n)*)$/;
	$call->{methodName}{_T} eq "pingback.ping"
		or return xmlerror("501 Not implemented", -32601,
						   "Method $call->{methodName}{_T} not supported");
	my @args = map { $_->{value}{string}{_T} || $_->{value}{_T} }
			   @{$call->{params}{param}};
	@args == 2
		or return xmlerror("400 Bad request", -32602, "Need 2 arguments");

	my ($src, $tgt) = @args;
	my ($sid) = ($tgt =~ m,^$baseurl/story/(\d+)$,)
		or return xmlerror("400 Bad request", 33, "Malformed target URI");
	my $stmt = $dbh->prepare_cached
		(qq{ SELECT count(*) FROM stories WHERE id = ? });
	$dbh->selectall_arrayref($stmt, {}, $sid)->[0]->[0] == 1
		or return xmlerror("400 Bad request", 33, "Target URI does not exist");
	
	my $ua = new LWP::UserAgent();
    my $r = new HTTP::Request("GET", $src);
    my $resp = $ua->request($r);
	$resp->is_success()
		or return xmlerror("400 Bad request", 16, "Cannot retrieve source URI");
	$resp->content() =~ /href="\Q$tgt\E"/s
		or return xmlerror("400 Bad request", 17, "Source URI does not refer to target URI");
	my ($title) = ($resp->content() =~ m,<title>\s*((.|\n)*?)\s*</title>,);

	$dbh->begin_work();
	$stmt = $dbh->prepare_cached
		(qq{ SELECT count(*) FROM pingbacks WHERE sid = ? AND link = ? });
	if ($dbh->selectall_arrayref($stmt, {}, $sid, $src)->[0]->[0] != 0) {
		$dbh->rollback();
		return xmlerror("400 Bad request", 48, "Pingback already registered");
	}
	if (!$dbh->do(qq{ INSERT INTO pingbacks (sid,date,link,title)
					  VALUES (?,CURRENT_TIMESTAMP,?,?) },
				  {}, $sid, $src, encode_utf8($title))) {
		$dbh->rollback();
		return xmlerror("400 Bad request", 0, "Database error");
	}
	$dbh->commit();

	vmsg("Received pingback from ".$req->connection()->remote_ip().
		 " for story id $sid");
	$req->status_line("200 Pingback added");
	$req->content_type("text/xml; charset=utf-8");
	print <<"EOF";
<?xml version="1.0" ?>
<methodResponse><params><param><value>
  <string>Pingback added</string>
</value></param></params></methodResponse>
EOF
	return OK;
}


sub xmlmember {
	my ($members, $name) = @_;
	my $x = (grep { $_->{name}{_T} eq $name } @$members)[0];
	return "" if !ref($x) eq "HASH";
	return(exists($x->{value}{boolean}{_T}) ? $x->{value}{boolean}{_T} :
		   exists($x->{value}{string}{_T}) ? $x->{value}{string}{_T} :
		   $x->{value}{_T});
}

