# Util.pm -- various utility functions for RoBlog
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
# more details.
# 
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307  USA.
#
package RoBlog::Util;

use Encode qw(_utf8_on);
use DBI;
use Apache2::Const;
use Apache2::Log;
use Apache::DBI;
use HTML::Entities;
use MIME::Base64;
use Digest::SHA1 qw(sha1 sha1_hex);
use XML::Parser;
use Time::Local;
use Time::Zone;
use RoBlog::Tmpl;
use RoBlog::Config;

use Exporter;
@ISA = qw(Exporter);
@EXPORT =
	qw(open_database dbget_mm dbget_1m dbget_m1 dbget_11
	   subst isodate shortdate excerpt summary comment_cred
	   atomfeedlink atompostlink atomeditlink atomtext
	   text_preprocess text_postprocess
	   vmsg showerror htmlerror xmlerror
	   inet_atoi inet_match wsse_check xmlparse);

use strict;
use vars qw($xmlparser $dbh $req $blockelts);
no warnings 'redefine';

$blockelts = "(address|blockquote|dl|dir|div|h[1-6]|hr|li|ol|p|pre|table|ul)";


sub open_database {
	$req = shift;
	$dbh = DBI->connect($dbi_uri, $db_user, $db_pass,
						{ PrintError => 0,
						  RaiseError => 1,
						  HandleError => sub { showerror($_[0]); },
						  AutoCommit => 1 })
		or showerror("Can't connect to database:\n$DBI::errstr\n");
	return $dbh;
}

sub utfdeca {
	my $a = shift or return undef;
	foreach my $h (@$a) {
		foreach (@$h) { _utf8_on($_); }
	}
	return $a;
}

sub utfdech {
	my $a = shift or return undef;
	foreach my $h (@$a) {
		foreach (keys %$h) { _utf8_on($h->{$_}); }
	}
	return $a;
}

sub dbget_mm {
	my ($query, @vals) = @_;
	my $stmt = $dbh->prepare_cached($query);
	return utfdech($dbh->selectall_arrayref($stmt, {Slice=>{}}, @vals));
}

sub dbget_m1 {
	my ($query, @vals) = @_;
	my $stmt = $dbh->prepare_cached($query);
	my $rows = utfdeca($dbh->selectall_arrayref($stmt, {}, @vals));
	return [map { $_->[0] } @$rows];
}

sub dbget_1m {
	my ($query, @vals) = @_;
	my $stmt = $dbh->prepare_cached($query);
	my $rows = utfdech($dbh->selectall_arrayref($stmt, {Slice=>{}}, @vals));
	return !@$rows ? undef : $rows->[0];
}

sub dbget_11 {
	my ($query, @vals) = @_;
	my $stmt = $dbh->prepare_cached($query);
	my $rows = utfdeca($dbh->selectall_arrayref($stmt, {}, @vals));
	return !@$rows ? undef : $rows->[0]->[0];
}


sub subst {
	my ($t, %s) = @_;
	while(my($k,$v) = each %s) {
		$t =~ s/__${k}__/$v/sg;
	}
	return $t;
}

sub isodate {
	my $t = shift;
	my $offset = tz_local_offset() / 60;
	$t =~ s/^(\d{4}-\d{2}-\d{2}) (\d{2}:\d{2}:\d{2})/$1T$2/;
	$t .= sprintf("%s%02d%02d", ($offset >= 0) ? '+' : '-',
				  abs(int($offset/60)), $offset%60);
	return $t;
}

sub isodate2time {
	my $s = shift;
	return undef if $s !~ /^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(Z|[+-]\d{4})$/;
	my $tz = $7;
	my $t = timegm($6, $5, $4, $3, $2-1, $1-1900);
	$t += ($1 eq "+" ? -1 : 1)*($2*3600 + $3*60)
		if $tz =~ /^([+-])(\d{2}\d{2})$/;
	return $t;
}

sub shortdate {
	my $d = shift;
	$d =~ /^(\S+ \d{2}:\d{2}).*$/;
	return $1;
}

sub excerpt {
	my ($t, $s) = @_;
	$t =~ s/<[^>]*>/ /g;
	$t =~ s/\s+/ /g;
	if ((my $pos = index($t, $s)) >= 0) {
		$t = substr($t, $pos<=50 ? 0 : $pos-50, 100);
		$t =~ s/^\S+\s+//;
		$t =~ s/\s+\S+$//;
		$t =~ s/\Q$s\E/subst($RoBlog::Tmpl::searchhighlight, TEXT=>$&)/eg;
	}
	return $t;
}

# generate a credential for a comment link, to avoid comment spam; it's not
# perfect, but should avoid most obvious abuses
# the credential is a checksum over the URL, plus a salt ($comment_cookie) to
# make it not trivially computable, plus current date to limit livetime
sub comment_cred {
	my @t = localtime;
	return sha1_hex($comment_cookie.
					sprintf("%03d%02d%02d", $t[5], $t[4], $t[3]).
					$_[0]);
}

sub atomfeedlink {
	return subst($RoBlog::Tmpl::atomlink,
				 REL => "feed", LINK => "feed");
}
sub atompostlink {
	return subst($RoBlog::Tmpl::atomlink,
				 REL => "server.post", LINK => "atom");
}
sub atomeditlink {
	return subst($RoBlog::Tmpl::atomlink,
				 REL => "service.edit", LINK => "atom\@$_[0]");
}

sub atomtext {
	my ($type, $text) = @_;

	if ($type =~ /x(ht)?ml/) {
		return($text =~ m,^\s*<div[^>]*>\s*((.|\n)*?)\s*</div>\s*$, ?
			   $1 : "ERROR no div tag around xhtml text");
	}
	elsif ($type =~ /^text/) {
		decode_entities($text);
		return $text;
	}
	elsif ($type =~ /html/) {
		decode_entities($text);
		decode_entities($text);
		return $text;
	}
	return "ERROR Unrecognized content type";
}

sub add_image {
	my ($fmt, $data) = @_;
	my $id;
	
	$dbh->begin_work();
	do {
		$id = int(rand(999999))+1;
	} while(dbget_11(qq{ SELECT id FROM images WHERE id=? }, $id));
	$data = decode_base64($data);
	$data =~ s/([\000-\037'\\\177-\377])/sprintf("\\%03o",ord($1))/ge;
	$dbh->do("INSERT INTO images (id,fmt,pic) VALUES (?,?,?)",
			 {}, $id, $fmt, $data);
	$dbh->commit();
	return $id;
}

sub process_linetag {
	my $t = shift;

	if ($t =~ /^(\w+)(:([^\]]+))?$/) {
		my ($cmd, $arg) = ($1, $3);
		my @args = split(/:/, $arg);
		if ($cmd eq "head") {
			$t = subst($RoBlog::Tmpl::bodyhead, TEXT => $args[0]);
		}
		elsif ($cmd eq "cat" || $cmd eq "title" || $cmd eq "date") {
			return "";
		}
		elsif ($cmd eq "img") {
			my $id = add_image($args[2], $args[3]);
			$t = subst($RoBlog::Tmpl::bodyimg{$args[1]}||$RoBlog::Tmpl::bodyimg,
					   LINK => "$basepath/pic/$id", NAME => $args[0]);
		}
		elsif ($cmd eq "clear") {
			$t = $RoBlog::Tmpl::bodyclear;
		}
		elsif ($cmd =~ /^([uo])list$/) {
			$t = $RoBlog::Tmpl::bodylist{"$1_start"}.
				 $RoBlog::Tmpl::bodylist{"item_start"};
		}
		elsif ($cmd =~ /^([uo])listend$/) {
			$t = $RoBlog::Tmpl::bodylist{"item_end"}.
				 $RoBlog::Tmpl::bodylist{"$1_end"};
		}
		elsif ($cmd eq "item") {
			$t = $RoBlog::Tmpl::bodylist{"item_end"}.
				 $RoBlog::Tmpl::bodylist{"item_start"};
		}
		elsif ($cmd eq "quote") {
			$t = $RoBlog::Tmpl::bodyquote;
		}
		elsif ($cmd eq "quoteend") {
			$t = $RoBlog::Tmpl::bodyquoteend;
		}
	}
	$t =~ s/\n+$//;
	return "\n$t\n\n";
}

sub text_preprocess {
	my ($t, $old) = @_;
	my @pingbacks;

	$t =~ s/\r//g;
	$t =~ s/^\[([^\]]*)\]\n+/process_linetag($1)/mge;
	$t =~ s/<p[^>]*>//g;
	$t =~ s/<\/p*>/\n\n/g;
	$t =~ s,\\$,<br/>,mg;
	while($t =~ m,\[(link:([^\]]+)\])?\[(https?:/[^\]]+)\],) {
		my ($name,$link) = ($2, $3);
		$name ||= $link;
		my $html = subst($RoBlog::Tmpl::bodylink, LINK => $link,TEXT => $name);
		$t =~ s{\[(link:([^\]]+)\])?\[(https?:/[^\]]+)\]}{$html};
		push(@pingbacks, $link)
			if !$old || $old !~ /<a[^>]*href="\Q$link\E"/;
	}
	while($t =~ /\[img:([^\]]+)\]/) {
		my @args = split(/:/, $1);
		my $id = add_image($args[2], $args[3]);
		my $html = subst($RoBlog::Tmpl::bodyimg{inline},
						 LINK => "$basepath/pic/$id", NAME => $args[0]);
		$t =~ s{\[img:[^\]]+\]}{$html};
	}
	$t =~ s/[ \t]+/ /g;
	$t =~ s/\n{2,}/\n\n/g;
	return ($t, \@pingbacks);
}

sub text_postprocess {
	my ($t) = @_;
	return join("\n", map { /^<\/?$blockelts[^>]*>/ ? $_ :
								subst($RoBlog::Tmpl::bodypara, TEXT => $_)
						  } split(/\n{2,}/, $t));
}


sub vmsg {
	my $t = shift;
	$req->log->notice($t) if $logging;
	return "$t\n";
}

sub showerror {
	my $txt = join("", @_);
	(my $status = "400 Internal error: $txt") =~ s/\n/ /mg;
	$req->status_line($status);
	print $txt;
	die "$txt\n";
}

sub htmlerror {
	my ($status, $text) = @_;
	$status =~ s/\n/ /mg;
	$req->status_line($status);
	$req->content_type("text/plain");
	print "$status\n$text\n";
	return OK;
}

sub xmlerror {
	my ($req, $status, $code, $text) = @_;
	vmsg($text);
	$req->status_line($status);
	$req->content_type("text/xml");
	print <<EOF;
<?xml version="1.0" ?>
<methodResponse><fault><value><struct>
 <member><name>faultCode</name><value><int>$code</int></value></member>
 <member><name>faultString</name><value><string>$text</string></value></member>
</struct></value></fault></methodResponse>
EOF
	return OK;
}

sub inet_atoi {
	my $str = shift;

	return ($1<<24)|($2<<16)|($3<<8)|$4
		if $str =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
	my ($ip6, $i) = ([], 0);
	$ip6->[$i++] = hex($1) while($str =~ s/^([[:xdigit:]]{1,4})://g);
	$i = 7;
	$ip6->[$i--] = hex($1) while($str =~ s/:([[:xdigit:]]{1,4})$//g);
	return $str =~ /^:?$/ ? $ip6 : 0;
}

sub inet_match {
	my ($ip1, $ip2, $len) = @_;
	my ($mask, $l);

	return 0 if ref($ip1) ne ref($ip2);
	if (ref($ip1)) {
		foreach my $i (0..7) {
			$l = $len > 16 ? 16 : $len <= 0 ? 0 : $len;
			$mask = (~((1 << (16-$l))-1)) & 0xffff;
			return 0 if ($ip1->[$i] & $mask) != ($ip2->[$i] & $mask);
			$len -= $l;
		}
		return 1;
	}
	else {
		$mask = (~((1 << (32-$len))-1)) & 0xffffffff;
		return ($ip1 & $mask) == ($ip2 & $mask);
	}
}

sub wsse_check {
	my ($wsse) = @_;
	my ($username, $pwdigest, $nonce, $date);
	$username = $1 if $wsse =~ /Username=\"([^\"]+)\"/mi; #"
	$pwdigest = $1 if $wsse =~ /PasswordDigest=\"([^\"]+)\"/mi; #"
	$nonce    = $1 if $wsse =~ /Nonce=\"([^\"]+)\"/mi; #"
	$date     = $1 if $wsse =~ /Created=\"([^\"]+)\"/mi; #"
	$nonce = decode_base64($nonce) if $nonce !~ /^[0-9a-fA-F]+$/;
	return 0 if !$username || !$pwdigest || !$nonce || !$date;
	$dbh->begin_work();
	$dbh->do(qq{ DELETE FROM used_nonces
				 WHERE age(current_timestamp,date) > interval '30 days' }, {});
	return 0 if dbget_11(qq{ SELECT count(*) FROM used_nonces
							 WHERE nonce = ? }, $nonce) > 0;
	$dbh->do(qq{ INSERT INTO used_nonces (nonce) VALUES (?) }, {}, $nonce);
	$dbh->commit();
	return 0 if $pwdigest ne encode_base64(sha1($nonce.$date.$blog_pwd), "");
	$date = isodate2time($date);
	my $now = time();
	return 0 if $date > $now || ($now-$date) > 60;
	return 1;
}

sub xmlparse {
	my ($text, $roottag, $texttags, $multitags) = @_;
	
	$xmlparser = new XML::Parser
		(Handlers => {
			Init => sub {
				my $xp = shift;
				$xp->{verb} = 0; $xp->{stack} = [];
				$xp->{cur} = $xp->{Tree} = {};
			},
			Final => sub {
				my $xp = shift;
				delete $xp->{cur}; delete $xp->{stack};
				$xp->{Tree};
			},
			Start => sub {
				my ($xp, $tag, %a) = @_;
				$xp->{cur}{_T} .= $xp->original_string(), return if $xp->{verb};
				my $newnode = { _A => {}, _T => "" };
				while(my($name,$val) = each %a) {
					$newnode->{_A}{$name} = $val;
				}
				$xp->{verb} = 1 if grep { $tag eq $_ } @{$ {$xp}{TextTags}};
				if ($ {$xp}{MultiTags} eq "ANY" ||
					grep { $tag eq $_ } @{$ {$xp}{MultiTags}}) {
					$xp->{cur}{$tag} ||= [];
					push(@{$xp->{cur}{$tag}}, $newnode);
				}
				else {
					$xp->xpcroak("Single tag <$tag> found more than once")
						if exists($xp->{cur}{$tag});
					$xp->{cur}{$tag} = $newnode;
				}
		 		push(@{$xp->{stack}}, $xp->{cur});
		 		$xp->{cur} = $newnode;
	 		},
			End => sub {
				my $xp = shift;
				if ($xp->{verb}) {
					$xp->{cur}{_T} .= $xp->original_string(), return
						if $xp->depth() >= @{$xp->{stack}};
					$xp->{verb} = 0;
				}
				delete $xp->{cur}{_T} if $xp->{cur}{_T} =~ /^(\s|\n)*$/;
				$xp->{cur} = pop(@{$xp->{stack}});
			},
			Char => sub {
				my ($xp, $text) = @_;
				$xp->{cur}{_T} .= $xp->{verb} ? $xp->original_string() : $text;
			}})
		if !defined $xmlparser;

	my $tree;
	eval {
		$tree = $xmlparser->parse($text, TextTags => $texttags,
								  MultiTags => $multitags);
	};
	return "ERROR $@" if $@;
	return "ERROR root tag is not $roottag" if !$tree->{$roottag};
	return $tree->{$roottag};
}

1;
